import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'package:logging/logging.dart';
import 'package:retry/retry.dart';
import 'package:risecx_chat_client/models/configuration_model.dart';
import 'package:risecx_chat_client/models/models.dart';
import 'package:risecx_chat_client/repositories/repositories.dart';
import 'package:rxdart/rxdart.dart';
import 'package:rxdart/subjects.dart';
import 'package:websocket_manager/websocket_manager.dart';

enum WS_CONNECTION_STATUS { DISCONNECTED, CONNECTED, CONNECTING, RECONNECTING }

class ChatBloc {
  WebsocketManager socket;
  List<dynamic> messages = [];
  dynamic selectedMessage;
  List<dynamic> favoriteMessages = [];
  List<dynamic> programs = [];
  List<QuickMenuItemModel> quickMenuItems = [];
  // bool userIdentified = false;
  Timer idleWatchTimer;
  final _chatRepository = ChatRepository();
  final _selectedMessageController = BehaviorSubject();
  final _messagesListController = BehaviorSubject(seedValue: []);
  final _favoriteMessagesListController = new BehaviorSubject(seedValue: []);
  final _programsListController = new BehaviorSubject(seedValue: []);
  final _quickMenuListController = new BehaviorSubject(seedValue: []);
  final _connectionStatusController = new BehaviorSubject();
  final Logger log = new Logger('ChatBloc');
  final ConfigurationModel configurationModel;
  final UserModel loggedUser;

  ChatBloc(this.configurationModel, this.loggedUser) {
    socket = WebsocketManager(this.configurationModel.wsUrl);
    connectToClient();
    __wsSetHandlers();
    __timeoutWatcher();
  }

  Observable get selecteMessagesToggleStream =>
      _selectedMessageController.stream;
  Observable get messagesListStream => _messagesListController.stream;

  Observable get favoriteMessagesListStream =>
      _favoriteMessagesListController.stream;

  Observable get programsListStream => _programsListController.stream;

  Observable get quickMenuListStream => _quickMenuListController.stream;

  Observable get wsConnectionStatusStream => _connectionStatusController.stream;

  void getHistory(UserModel user) async {
    try {
      if (messages.isEmpty) {
        log.fine('Messages has ${messages.length} items... recovering history');
        ChatHistortModel chatHistortModel =
            await _chatRepository.getHistory(user);
        messages.addAll(_historyMessagesHandler(chatHistortModel));
        _messagesListController.add(messages);
      }
    } catch (e) {
      log.severe('getHistory Fail!', e);
    }
  }

  void getFavoriteMessages(UserModel user) async {
    try {
      List<dynamic> favorites = await _chatRepository.getFavorites(user);
      favorites.forEach((favorite) {
        if (!favoriteMessages.any((f) => f['_id'] == favorite['_id'])) {
          favorite['message']['_id'] = favorite['_id'];
          favoriteMessages.add(favorite);
        }
      });
      _favoriteMessagesListController.add(favoriteMessages);
    } catch (e) {
      log.severe('getFavoriteMessages Fail!', e);
      throw e;
    }
  }

  void getPrograms(UserModel user) async {
    try {
      List<dynamic> _programs = await _chatRepository.getPrograms(user);
      _programs.forEach((program) {
        if (!programs.any((p) => p['id'] == program['id'])) {
          programs.add(program);
        }
      });
      _programsListController.add(programs);
    } catch (e) {
      throw e;
    }
  }

  void addOrRemoveFavoriteMessage(dynamic message, bool favorite) async {
    try {
      final result =
          await _chatRepository.addOrRemoveFavoriteMessage(message, favorite);
      if (!favorite) {
        favoriteMessages.removeWhere((f) => f['_id'] == result['_id']);
      } else {
        result['message']['_id'] = result['_id'];
        favoriteMessages.add(result);
      }
      selectMessageToggle(result);
      _favoriteMessagesListController.add(favoriteMessages);
    } catch (e) {
      log.severe('addOrRemoveFavoriteMessage Fail!', e);
      throw e;
    }
  }

  void subscribeUnsubscribe(ProgramModel program, UserModel user) async {
    try {
      final result = await _chatRepository.subscribeUnsubscribe(program, user);
      programs =
          programs.map((e) => e['_id'] == result['_id'] ? result : e).toList();

      _programsListController.add(programs);
    } catch (e) {
      log.severe('subscribeUnsubscribe Fail!', e);
      throw e;
    }
  }

  void getQuickMenuItems() async {
    try {
      Iterable quickMenuList = await _chatRepository.getQuickMenu();
      quickMenuItems =
          quickMenuList.map((qm) => QuickMenuItemModel.fromJson(qm)).toList();
      _quickMenuListController.add(quickMenuItems);
    } catch (e) {
      log.severe('getQuickMenuItems Fail!', e);
      throw e;
    }
  }

  List<dynamic> addMessage(dynamic message) {
    if (message['type'] != 'typing' && message['type'] != 'identify') {
      messages.add(message);
      _messagesListController.add(messages);
    }
    if (idleWatchTimer != null) {
      idleWatchTimer.cancel();
      __timeoutWatcher();
    }
    return messages;
  }

  List<dynamic> _historyMessagesHandler(ChatHistortModel chatHistortModel) {
    List<dynamic> listHistoryMessages = new List<dynamic>();
    try {
      chatHistortModel.history.map((h) {
        if (h.message['raw_message'] != null) {
          h.message['text'] = h.message['raw_message']['payloadTitle'] != null
              ? h.message['raw_message']['payloadTitle']
              : h.message['raw_message']['text'];
        }
        h.message['_id'] = h.sId;
        h.message['favorite'] = h.favorite;
        return h;
      }).forEach((h) {
        listHistoryMessages.add(h.message);
      });
    } catch (e) {
      log.severe('_historyMessagesHandler fail!', e);
      throw e;
    }
    return listHistoryMessages;
  }

  void selectMessageToggle(dynamic message) {
    print(message);
    if (selectedMessage != null) {
      selectedMessage =
          message['_id'] == selectedMessage['_id'] ? null : message;
    } else {
      selectedMessage = message;
    }
    _selectedMessageController.add(selectedMessage);
  }

  void sendMessage(SendMessageModel message, {bool isPostback = false}) {
    final jsonMessage = jsonEncode(message);
    log.fine('sending message:${message.text}');
    if (!isPostback) socket.send(jsonMessage);
    addMessage(jsonDecode(jsonMessage));
  }

  void onPostBack(SendMessageModel message, title, payload) {
    sendMessage(message, isPostback: true);
    socket.send(jsonEncode({
      'channel': 'socket',
      'type': 'message',
      'user': message.user,
      'payloadTitle': title,
      'text': payload
    }));
  }

  void userIdentifier() {
    try {
      final SendMessageModel message = new SendMessageModel(
          user: loggedUser.id, type: 'identify', channel: 'socket');
      socket.send(jsonEncode(message));
      print('User ${loggedUser.sId} identified');
    } catch (e) {
      throw e;
    }
  }

  void connectToClient({bool reconnection = false}) async {
    try {
      final r = RetryOptions(maxAttempts: 8);
      reconnection
          ? _connectionStatusController.add(WS_CONNECTION_STATUS.RECONNECTING)
          : _connectionStatusController.add(WS_CONNECTION_STATUS.CONNECTING);
      ;
      await r.retry(() => socket.connect().timeout(Duration(seconds: 30)),
          retryIf: (e) => e is SocketException || e is TimeoutException,
          onRetry: (e) {
            print('connectToClient.onRetry $e');
            _connectionStatusController.add(WS_CONNECTION_STATUS.RECONNECTING);
          });
      userIdentifier();
    } finally {
      _connectionStatusController.add(WS_CONNECTION_STATUS.CONNECTED);
    }
  }

  void closeConnection() {
    socket.close();
  }

  void __wsSetHandlers() {
    if (socket != null) {
      socket.onMessage(__onMessageHandler);
      socket.onClose((dynamic message) {
        print('websocketBloc ====> Close $message');
        _connectionStatusController.add(WS_CONNECTION_STATUS.DISCONNECTED);
        Timer(Duration(seconds: 3), () => connectToClient(reconnection: true));
      });
    }
  }

  void __onMessageHandler(dynamic wsMessage) {
    print('WebsocketBloc recv: $wsMessage');
    addMessage(jsonDecode(wsMessage));
  }

  void onMediaView({String url, String blockId, String sessionId}) {
    _chatRepository.eventCounterReport(
        {'url': url, 'blockId': blockId, 'sessionId': sessionId});
  }

  void __timeoutWatcher() {
    Timer.periodic(Duration(minutes: 8), (timer) {
      idleWatchTimer = timer;
      print('__timeoutWatcher inativo por 8 min...');
      userIdentifier();
    });
  }

  dispose() {
    _selectedMessageController.close();
    _messagesListController.close();
  }
}
