import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'package:logging/logging.dart';
import 'package:retry/retry.dart';
import 'package:risecx_chat_client/models/configuration_model.dart';
import 'package:risecx_chat_client/models/models.dart';
import 'package:risecx_chat_client/repositories/repositories.dart';
import 'package:rxdart/rxdart.dart';
import 'package:rxdart/subjects.dart';
import 'package:web_socket_channel/io.dart';
import 'package:web_socket_channel/status.dart' as status;
import 'package:web_socket_channel/web_socket_channel.dart';

enum WS_CONNECTION_STATUS { DISCONNECTED, CONNECTED, CONNECTING, RECONNECTING }

class ChatBloc {
  IOWebSocketChannel socket;
  List<dynamic> messages = [];
  dynamic selectedMessage;
  List<dynamic> favoriteMessages = [];
  List<dynamic> programs = [];
  List<QuickMenuItemModel> quickMenuItems = [];
  bool historyWasRecovered = false;
  bool initialMessageSent = false;
  // bool userIdentified = false;
  Timer idleWatchTimer;
  final _chatRepository = ChatRepository();
  final _selectedMessageController = BehaviorSubject();
  final _messagesListController = BehaviorSubject(seedValue: []);
  final _favoriteMessagesListController = new BehaviorSubject(seedValue: []);
  final _programsListController = new BehaviorSubject(seedValue: []);
  final _quickMenuListController =
      new BehaviorSubject(seedValue: <QuickMenuItemModel>[]);
  final _connectionStatusController = new BehaviorSubject();
  final _userIdenfierStatusController = new BehaviorSubject(seedValue: false);
  final Logger log = new Logger('ChatBloc');
  final ConfigurationModel configurationModel;
  final UserModel loggedUser;

  ChatBloc(this.configurationModel, this.loggedUser) {
    connectToClient();
    __timeoutWatcher();
    _connectionStatusController.stream.listen((event) {
      if (event == WS_CONNECTION_STATUS.CONNECTED)
        Timer(Duration(seconds: 3), userIdentifier);
    });
  }

  Observable get selecteMessagesToggleStream =>
      _selectedMessageController.stream;
  Observable get messagesListStream => _messagesListController.stream;

  Observable get favoriteMessagesListStream =>
      _favoriteMessagesListController.stream;

  Observable get programsListStream => _programsListController.stream;

  Observable get quickMenuListStream => _quickMenuListController.stream;

  Observable get wsConnectionStatusStream => _connectionStatusController.stream;

  Observable get userIdentifierStatusStream =>
      _userIdenfierStatusController.stream;

  Future getHistory(UserModel user) async {
    try {
      if (historyWasRecovered) return;
      ChatHistortModel chatHistortModel =
          await _chatRepository.getHistory(user);
      log.fine('Messages has ${messages.length} items... recovering history');
      messages.isEmpty
          ? messages.addAll(_historyMessagesHandler(chatHistortModel))
          : messages.insertAll(0, _historyMessagesHandler(chatHistortModel));
      _messagesListController.add(messages);
      historyWasRecovered = true;
    } catch (e) {
      print('getHistory Fail! $e');
    }
  }

  void getFavoriteMessages(UserModel user) async {
    try {
      List<dynamic> favorites = await _chatRepository.getFavorites(user);
      favorites.forEach((favorite) {
        if (!favoriteMessages.any((f) => f['_id'] == favorite['_id'])) {
          favorite['message']['_id'] = favorite['_id'];
          favoriteMessages.add(favorite);
        }
      });
      _favoriteMessagesListController.add(favoriteMessages);
    } catch (e) {
      log.shout('getFavoriteMessages Fail!', e);
      throw e;
    }
  }

  void getPrograms(UserModel user) async {
    try {
      List<dynamic> _programs = await _chatRepository.getPrograms(user);
      _programs.forEach((program) {
        if (!programs.any((p) => p['id'] == program['id'])) {
          programs.add(program);
        }
      });
      _programsListController.add(programs);
    } catch (e) {
      throw e;
    }
  }

  void addOrRemoveFavoriteMessage(dynamic message, bool favorite) async {
    try {
      final result =
          await _chatRepository.addOrRemoveFavoriteMessage(message, favorite);
      if (!favorite) {
        favoriteMessages.removeWhere((f) => f['_id'] == result['_id']);
      } else {
        result['message']['_id'] = result['_id'];
        favoriteMessages.add(result);
      }
      selectMessageToggle(result);
      _favoriteMessagesListController.add(favoriteMessages);
    } catch (e) {
      log.severe('addOrRemoveFavoriteMessage Fail!', e);
      throw e;
    }
  }

  void subscribeUnsubscribe(ProgramModel program, UserModel user) async {
    try {
      final result = await _chatRepository.subscribeUnsubscribe(program, user);
      programs =
          programs.map((e) => e['_id'] == result['_id'] ? result : e).toList();

      _programsListController.add(programs);
    } catch (e) {
      log.severe('subscribeUnsubscribe Fail!', e);
      throw e;
    }
  }

  void getQuickMenuItems() async {
    try {
      Iterable quickMenuList = await _chatRepository.getQuickMenu();
      quickMenuItems =
          quickMenuList.map((qm) => QuickMenuItemModel.fromJson(qm)).toList();
      _quickMenuListController.add(quickMenuItems);
    } catch (e) {
      log.severe('getQuickMenuItems Fail!', e);
      throw e;
    }
  }

  List<dynamic> addMessage(dynamic message) {
    if (message['type'] != 'identify') {
      messages.add(message);
      _messagesListController.add(messages);
    }
    return messages;
  }

  List<dynamic> _historyMessagesHandler(ChatHistortModel chatHistortModel) {
    List<dynamic> listHistoryMessages = new List<dynamic>();
    try {
      chatHistortModel.history.map((h) {
        h.message['author'] = h.message['to'] != null ? 'them' : 'me';
        h.message['_id'] = h.sId;
        h.message['favorite'] = h.favorite;
        return h;
      }).forEach((h) {
        listHistoryMessages.add(h.message);
      });
    } catch (e) {
      log.severe('_historyMessagesHandler fail!', e);
      throw e;
    }
    return listHistoryMessages;
  }

  void selectMessageToggle(dynamic message) {
    print(message);
    if (selectedMessage != null) {
      selectedMessage =
          message['_id'] == selectedMessage['_id'] ? null : message;
    } else {
      selectedMessage = message;
    }
    _selectedMessageController.add(selectedMessage);
  }

  void sendMessage(SendMessageModel message,
      {bool isPostback = false, bool initialMessage = false}) {
    final jsonMessage = jsonEncode(message);
    print('sending message:${message.text}');
    if (!isPostback) socket.sink.add(jsonMessage);
    addMessage(jsonDecode(jsonMessage));
    if (initialMessage) initialMessageSent = true;
  }

  void onPostBack(SendMessageModel message, title, payload,
      {bool initialMessage = false}) {
    sendMessage(message, isPostback: true, initialMessage: initialMessage);
    socket.sink.add(jsonEncode({
      'channel': 'socket',
      'type': 'message',
      'user': message.user,
      'payloadTitle': title,
      'text': payload
    }));
  }

  void userIdentifier() {
    try {
      final SendMessageModel message = new SendMessageModel(
          user: loggedUser.id, type: 'identify', channel: 'socket');
      socket.sink.add(jsonEncode(message));
      print('User ${loggedUser.sId} identified');
      _userIdenfierStatusController.add(true);
    } catch (e) {
      throw e;
    }
  }

  void connectToClient({bool reconnection = false}) async {
    try {
      final r = RetryOptions(maxAttempts: 8);
      reconnection
          ? _connectionStatusController.add(WS_CONNECTION_STATUS.RECONNECTING)
          : _connectionStatusController.add(WS_CONNECTION_STATUS.CONNECTING);

      await r.retry(
          () {
            socket = IOWebSocketChannel.connect(
              this.configurationModel.wsUrl,
            );
            __wsSetHandlers(socket);
          },
          retryIf: (e) =>
              e is SocketException ||
              e is TimeoutException ||
              e is WebSocketChannelException,
          onRetry: (e) {
            print('connectToClient.onRetry $e');
            _connectionStatusController.add(WS_CONNECTION_STATUS.RECONNECTING);
          });
    } finally {
      _connectionStatusController.add(WS_CONNECTION_STATUS.CONNECTED);
    }
  }

  void closeConnection() {
    socket.sink.close(status.goingAway);
  }

  void __wsSetHandlers(IOWebSocketChannel socket) {
    socket.stream.listen(__onMessageHandler,
        onDone: () {
          print('websocketBloc ====> Close ');
          _connectionStatusController.add(WS_CONNECTION_STATUS.DISCONNECTED);
          connectToClient();
        },
        onError: (e) => print('websocketBloc ====> Error $e'));
  }

  void __onMessageHandler(dynamic wsMessage) {
    print('WebsocketBloc recv: $wsMessage');
    addMessage(jsonDecode(wsMessage));
  }

  void onMediaView({String url, String blockId, String sessionId}) {
    _chatRepository.eventCounterReport(
        {'url': url, 'blockId': blockId, 'sessionId': sessionId});
  }

  void __timeoutWatcher() {
    Timer.periodic(Duration(minutes: 8), (timer) {
      idleWatchTimer = timer;
      print('Keep Alive Sir');
      userIdentifier();
    });
  }

  dispose() {
    _selectedMessageController.close();
    _messagesListController.close();
    closeConnection();
  }
}
